/************************************************************************/
/* aux: auxiliary graphics subroutines					*/
/************************************************************************/
#include "../include/xml2eps.h"


/* global variables */

extern int actfontsize, actfont;
extern const char *families[], *fonts[];


/************************************************************************/
/* function: math interpreter for user-defined function			*/
/************************************************************************/
/* x		(i)  argument
   o		(i)  auxiliary argument
   par		(i)  parameter vector
   op		(i)  string defining the function (reverse Polish notation)
   y		(o)  function value
   RESULT	(o)  error code: 0 = no error
				 1 = stack empty
				 2 = stack overflow
				 3 = operator not recognized
				 4 = operation failed (e.g., inv of 0)	*/
/*======================================================================*/
int function(double x, double o, const double *par, const char *op,
double *y) {
  int level, i, k;
  char *p, opcode[GRX_RLEN];
  double stack[GRX_MSTACK], *x0, *x1;

  level = 0;
  stack[0] = 0.0;
  p = (char *) op;
  while (*p) {
    i = sscanf(p, "%[^,]%n", opcode, &k);
    if (i < 1) break;
    p += k;
    if (*p != '\0') p++;		/* skip over the comma */

    x0 = &stack[level - 1];		/* uppermost stack element */
    x1 = x0 - 1;			/* next one */

    if (!strcmp(opcode, "+")) {
      if (level < 2) return 1;
      *x1 += *x0;
      level--;
    }
    else if (!strcmp(opcode, "-")) {
      if (level < 2) return 1;
      *x1 -= *x0;
      level--;
    }
    else if (!strcmp(opcode, "*")) {
      if (level < 2) return 1;
      *x1 *= *x0;
      level--;
    }
    else if (!strcmp(opcode, "/")) {
      if (level < 2) return 1;
      *x1 /= *x0;
      level--;
    }
    else if (!strcmp(opcode, "^")) {
      if (level < 2) return 1;
      *x1 = pow(fabs(*x1), *x0);
      level--;
    }
    else if (!strcmp(opcode, "abs")) {
      if (level < 1) return 1;
      *x0 = fabs(*x0);
    }
    else if (!strcmp(opcode, "cos")) {
      if (level < 1) return 1;
      *x0 = cos(*x0);
    }
    else if (!strcmp(opcode, "dup")) {	/* duplicate uppermost stack */
      if (level == GRX_MSTACK) return 2;
      stack[level] = *x0;
      level++;
    }
    else if (!strcmp(opcode, "erf")) {
      if (level < 1) return 1;
      *x0 = erf(*x0);
    }
    else if (!strcmp(opcode, "erfc")) {
      if (level < 1) return 1;
      *x0 = erfc(*x0);
    }
    else if (!strcmp(opcode, "exp")) {
      if (level < 1) return 1;
      *x0 = exp(*x0);
    }
    else if (!strcmp(opcode, "inv")) {
      if (level < 1) return 1;
      if (*x0 != 0.0)
	*x0 = 1.0 / *x0;
      else
	return 4;
    }
    else if (!strcmp(opcode, "lg")) {
      if (level < 1) return 1;
      if (*x0 > 0.0)
	*x0 = log10(*x0);
      else
	return 4;
    }
    else if (!strcmp(opcode, "ln")) {
      if (level < 1) return 1;
      if (*x0 > 0.0)
	*x0 = log(*x0);
      else
	return 4;
    }
    else if (!strcmp(opcode, "neg")) {
      if (level < 1) return 1;
      *x0 = -*x0;
    }
    else if (!strcmp(opcode, "o")) {
      if (level == GRX_MSTACK) return 2;
      stack[level] = o;
      level++;
    }
    else if (!strcmp(opcode, "sin")) {
      if (level < 1) return 1;
      *x0 = sin(*x0);
    }
    else if (!strcmp(opcode, "tan")) {
      if (level < 1) return 1;
      *x0 = tan(*x0);
    }
    else if (!strcmp(opcode, "x")) {
      if (level == GRX_MSTACK) return 2;
      stack[level] = x;
      level++;
    }
    else if (*opcode == 'p') {		/* parameter */
      if (level == GRX_MSTACK) return 2;
      i = sscanf(opcode + 1, "%d", &k);
      if (i == 1) {
	stack[level] = par[k];
	level++;
      }
      else
	return 3;
    }
    else {				/* has got to be a number */
      if (level == GRX_MSTACK) return 2;
      i = sscanf(opcode, "%lf", &stack[level]);
      if (i == 1)
	level++;
      else
	return 3;
    }
  }
  *y = stack[level - 1];
  return GRX_GOOD;
}

/************************************************************************/
/* math_op: mathematical operations on a curve data array		*/
/************************************************************************/
/* x		(io) data array
   n		(i)  its length
   xh		(i)  another data array of the same length
   par		(i)  parameters for function
   op		(i)  RPN operation string (syntax: see function() )
   RESULT	(o)  error code: 0 = no error
				 1 = stack empty
				 2 = stack overflow
				 3 = operator not recognized
				 4 = operation failed (e.g., inv of 0)	*/
/*======================================================================*/
int math_op(double *x, int n, const double *xh, const double *par,
const char *op) {
  int err;
  double o, *xmax;

  err = 0;
  if (*op != '\0') {
    xmax = x + n;
    while (x < xmax) {
      o = (xh == NULL) ? 0.0 : *xh++;
      if ((err = function(*x, o, par, op, x)) != GRX_GOOD) break;
      x++;
    }
  }
  return err;
}

/************************************************************************/
/* transform_xy: apply mathematical operations to an (x, y) pair	*/
/************************************************************************/
/* x, y		(i)  x, y input coordinates
   par		(i)  parameters (note: these parameters will be used by
   		     the x AND y transformations)
   xop, yop	(i)  RPN operation strings (syntax: see function() )
   xt, yt	(o)  transformed coordinates
   RESULT	(o)  error code: 0 = no error				*/
/*======================================================================*/
int transform_xy(double x, double y, const double *par, const char *xop,
const char *yop, double *xt, double *yt) {
  int xerr, yerr;

  xerr = yerr = 0;
  if (*xop == '\0')
    *xt = x;
  else {
    xerr = function(x, y, par, xop, xt);
    if (xerr != 0) fprintf(stderr, "<transform_xy>: x transformation "
    "failed at x = %g with error %d\n", x, xerr);
  }
  if (*yop == '\0')
    *yt = y;
  else {
    yerr = function(y, x, par, yop, yt);
    if (yerr != 0) fprintf(stderr, "<transform_xy>: y transformation "
    "failed at y = %g with error %d\n", y, yerr);
  }
  return (xerr == 0 && yerr == 0) ? 0 : 1;
}

/************************************************************************/
/* optimize: fit parameters to data					*/
/************************************************************************/
/* x, y		(i)  data array
   n		(i)  number of data points
   par		(io) parameter array
   rpn		(i)  RPN string defining the fitting function
   fit		(i)  string naming the adjustable parameters
   RESULT	(o)  error code: 0 = no error
				 others see eqn_marquardt.c		*/
/*======================================================================*/
int optimize(const double *x, const double *y, int n, double *par,
const char *rpn, const char *fit) {
  int ifit[GRX_NPAR], i, err, m, k, j;
  double adjpar[GRX_NPAR], h;
  char *p;
  struct objaux aux;

  p = (char *) fit;			/* get indices of adj. parameters */
  m = 0;
  while ((i = sscanf(p, "%d%n", &j, &k)) > 0) {
    ifit[m] = j;
    p += k;
    while (*p == ',') p++;
    m++;
    if (m >= GRX_NPAR) {
      fprintf(stderr, "warning: optimize -  too many adjustable parameters\n");
      break;
    }
  }

  for (i = 0; i < m; i++) adjpar[i] = par[ifit[i]];
  aux.par = par;
  aux.funct = (char *) rpn;
  aux.x = (double *) x;
  aux.y = (double *) y;
  aux.n = n;
  aux.m = m;
  aux.ifit = ifit;

  if ((err = fit_marquardt(objfct, m, adjpar, n, &aux, 20 * m, 0, &h))
  != GRX_GOOD)
    fprintf(stderr, "warning: optimize - parameter fitting failed with "
    "error %d\n", err);
  else 
    fprintf(stderr, "report: optimize - parameter fitting succeeded\n");
  fprintf(stderr, "adjustable parameters changed to:\n");
  for (i = 0; i < m; i++) fprintf(stderr, "%2d %e\n", i, adjpar[i]);
  fprintf(stderr, "\nr.m.s. deviation: %e\n\n", h);
  return err;
}

/************************************************************************/
/* objfct: object function for optimize					*/
/************************************************************************/
/* adjpar	(i)  adjustable parameters
   r		(o)  residuals
   aux		(i)  auxiliary data
   RESULT	(o)  error code: 0 = no error				*/
/*======================================================================*/
int objfct(double *adjpar, double *r, struct objaux *aux) {
  int i, err;

  for (i = 0; i < aux->m; i++) aux->par[aux->ifit[i]] = adjpar[i];

  for (i = 0; i < aux->n; i++) {
    if ((err = function(aux->x[i], 0.0, aux->par, aux->funct, &r[i]))
    != GRX_GOOD) break;
    r[i] -= aux->y[i];
  }
  return err;
}

/************************************************************************/
/* split_unit: split unit off a TeX-style dimension			*/
/*	       (as a preparation for conversion to PostScript)		*/
/************************************************************************/
/* s		(i)  string containing a TeX-style dimension, e.g., 5mm
   v		(o)  value part of the input, e.g., 5
   u		(o)  unit part of the input, e.g., "mm"			*/
/*======================================================================*/
void split_unit(const char *s, double *v, char *u) {
  int i;

  sscanf(s, "%lf%n", v, &i);
  if (s[i] == '\0' || s[i] == ' ')	/* no unit - treat as relative */
    strcpy(u, "ru");
  else if (!strcmp(s + i, "pt"))	/* pt is intrinsic to PS: omit */
    *u = '\0';
  else
    strcpy(u, s + i);
}

/************************************************************************/
/* scale: calculation of scaling parameters and tic mark positions	*/
/************************************************************************/
/* smin, smax	(io) min. and max. coordinate value
   delta	(o)  distance of minor tic marks
   major	(o)  number of minor tics to each major			*/
/*======================================================================*/
void scale(double smin, double smax, double *delta, int *major) {
  int i;
  double pdelta, sdiff;

  sdiff = smax - smin;
  *delta = pow(10.0, ceil(log10(sdiff)));
  pdelta = 5 * (*delta);
  i = 0;
  while (sdiff <= 4 * pdelta) {
    switch (i % 4) {			/* minor  major */
    case 0:				/*  1.0    5.0 */
      *major = 2;
      break;
    case 1:				/*  1.0    2.0 */
      *delta *= 0.5;
      break;
    case 2:				/*  0.5    1.0 */
      *delta *= 0.4;
      *major = 5;
      break;
    case 3:				/*  0.2    1.0 */
      *delta *= 0.5;
    }
    pdelta = *major * *delta;
    i++;
  }
}

/************************************************************************/
/* set_color: set color of plotted objects				*/
/************************************************************************/
/* color	(i)  6-digit hexadecimal code for RGB, e.g., 0x0000ff
		     for blue, or a color name as listed in rgb.txt
   RESULT	(o)  error code: 0 = no error
				 1 = color not recognized (use default)	*/
/*======================================================================*/
int set_color(const char *color) {
  int err;
  unsigned int i, red, green, blue;
  char colorfile[GRX_PATHLENGTH], cname[GRX_ALEN];
  FILE *cols;

  red = green = blue = 0;		/* default color: black */

  if (strncmp(color, "0x", 2) == 0) {	/* color specified as hexcode */
    err = sscanf(color, "%x", &i);
    if (err == 0 || i > 0xffffff)
      err = 1;
    else {
      err = 0;
      blue = i & 0xff;
      green = (i >> 8) & 0xff;
      red = (i >> 16) & 0xff;
    }
  }
  else {				/* color specified by name */
    sprintf(colorfile, "%s/xml/rgb.txt", GRX_HOME);
    if ((cols = fopen(colorfile, "r")) == NULL)
      err = 2;
    else {
      while ((i = fscanf(cols, "%u%u%u%s", &red, &green, &blue, cname))
      == 4) {
	if (!strcasecmp(color, cname)) break;
      }
      err = (i == 4) ? 0 : 3;
      fclose(cols);
    }
  }
  printf("  %f %f %f setrgbcolor\n", ((double) red )/255.0,
  ((double) green)/255.0, ((double) blue)/255.0);
  return err;
}

/************************************************************************/
/* set_font: put PostScript commands for font selection			*/
/************************************************************************/
/* family	(i)  font family name, e.g., Times-Roman
   weight	(i)  medium or bold
   shape	(i)  roman or italic
   size		(i)  size in pt
   RESULT	(o)  error code: 0 = no error
				 1 = unknown font (use default)		*/
/*======================================================================*/
int set_font(const char *family, const char *weight, const char *shape,
int size) {
  int k, err;

  if ((k = identify(family, families)) < 0) {
    k = 0;
    err = 1;
  }
  else
    err = 0;
  k *= 4;
  if (!strcmp(weight, "bold")) k += 2;
  if (!strcmp(shape, "italic")) k++;
  if (k != actfont || size != actfontsize) {
    actfont = k;
    actfontsize = size;
    printf("    /%s findfont %d scalefont setfont\n", fonts[k], size);
  }
  return err;
}

/************************************************************************/
/* set_line_style: set appearance of straight lines			*/
/************************************************************************/
/* width	(i)  thickness in pt
   pattern	(i)  line pattern
   RESULT	(o)  error code: 0 = no error
				 1 = line width wrong (use default)
				 2 = line pattern wrong	(use default	*/
/*======================================================================*/
int set_line_style(double width, int pattern) {
  int err;

  err = 0;
  if (width < 0.0) {
    width = 1.0;
    err = 1;
  }
  printf("  %f setlinewidth\n", width);

  switch (pattern) {
  case SOLID:
    printf("  []");
    break;
  case DASH:
    printf("  [6 6]");
    break;
  case DOT:
    printf("  [1 3]");
    break;
  case DASH_DOT:
    printf("  [7 4 1 4]");
    break;
  case DASH_DOT_DOT:
    printf("  [6 3 1 2 1 3]");
    break;
  case LDASH:
    printf("  [12 4]");
    break;
  case LDASH_DASH:
    printf("  [9 2 3 2]");
    break;
  case LDASH_DASH_DASH:
    printf("  [10 2 3 2 3 2]");
    break;
  default:
    printf("  []");
    err = 2;
  }
  printf(" 0 setdash\n");
  return err;
}

/************************************************************************/
/* print_text: print a simple text (without embedded SGML tags)		*/
/************************************************************************/
/* s		(i)  text to be printed
   x, y		(i)  coordinates in xu, yu
   xalign	(i)  x alignment
   yalign	(i)  y alignment
   fontfamily	(i)  e.g., Helvetica
   fontweight	(i)  medium or bold
   fontshape	(i)  roman or italic
   fontsize	(i)  in arbitrary units
   phi		(i)  rotation angle in degrees				*/
/*======================================================================*/
void print_text(const char *s, double x, double y, const char *xalign,
const char *yalign, const char *fontfamily, const char *fontweight,
const char *fontshape, const char *fontsize, double phi) {
  set_font(fontfamily, fontweight, fontshape, (int) psunits(fontsize));
  printf("\ngsave\n");
  printf("  %g xu %g yu translate 0 0 m\n", x, y);
  if (phi != 0.0) printf("  %g rotate\n", phi);
  printf("  (%s) dup\n", s);
  printf("  false charpath flattenpath pathbbox\n");
					/* stack: llx lly urx ury */
  printf("  exch 4 -1 roll sub ");	/* stack: lly ury width */
  if (!strcmp(xalign, "right"))
    printf("neg\n");
  else if (!strcmp(xalign, "center"))
    printf("-0.5 mul\n");
  else
    printf("pop 0\n");
  printf("  3 1 roll exch sub ");	/* stack: dx height */
  if (!strcmp(yalign, "ceiling"))
    printf("pop %g ", -1.2 * psunits(fontsize));
  else if (!strcmp(yalign, "top"))
    printf("neg ");
  else if (!strcmp(yalign, "center"))
    printf("-0.5 mul ");
  else
    printf("pop 0 ");
  printf("0 0 m M show\n");
  printf("grestore\n");
}

/************************************************************************/
/* curve_length: estimate the lengths of the curve segments between	*/
/*	         data points (needed for parametric interpolations)	*/
/************************************************************************/
/* n		(i)  number of data points
   x, y		(i)  coordinate arrays
   mode		(i)  estimation mode: 0 = linear segment between points
				      1 = circle segment through 3 points
   t		(o)  array of cumulated curve lengths			*/
/*======================================================================*/
void curve_length(int n, double *x, double *y, int mode, double *t) {
  int i;
  double a[2][2], b[2], xm, ym, det, r, dt, scale;

  scale = psunits("1xu")/psunits("1yu");/* if the units are very different */

  for (i = 1, t[0] = 0.0; i < n; i++) {
    dt = sqrt(SQ(x[i] - x[i-1]) + SQ(scale * (y[i] - y[i-1])));
    if (mode && i >= 2) {
      a[0][0] = 2 * (x[i-1] - x[i]);	/* set up eqn. for center */
      a[0][1] = 2 * (y[i-1] - y[i]) * scale;
      a[1][0] = 2 * (x[i-2] - x[i]);
      a[1][1] = 2 * (y[i-2] - y[i]) * scale;
      b[0] = SQ(x[i-1]) - SQ(x[i]) + (SQ(y[i-1]) - SQ(y[i])) * SQ(scale);
      b[1] = SQ(x[i-2]) - SQ(x[i]) + (SQ(y[i-2]) - SQ(y[i])) * SQ(scale);
      det = a[0][0] * a[1][1] - a[0][1] * a[1][0];
      if (det != 0.0) {
	xm = (b[0] * a[1][1] - b[1] * a[0][1]) / det;
	ym = (b[0] * a[1][0] - b[1] * a[0][0]) / det;
	r = sqrt(SQ(x[i] - xm) + SQ(scale * y[i] - ym));
        dt = 2 * r * asin(0.5 * dt / r);
      }
    }
    t[i] = t[i-1] + dt;
  }
}
