/************************************************************************/
/* xml2eps: simple and portable XML to Encapsulated PostScript		*/
/*	    converter (for diagrams)					*/
/************************************************************************/
/* I/O channels
   cmd line	Grafix diagram directory
   xml:	XML diagram file
   style:	associated style file (global style settings)
   stdout	EPS output
   stderr	error messages						*/
/*======================================================================*/
#include "../include/xml2eps.h"
#include "../include/keywords.h"


/* global variables */

int
  interpretation_mode = XML,
  actfontsize = 0,
  actfont = -1,
  radix_style = 0;			/* use decimal dot */
double par[GRX_NPAR];
struct attlist diagram,			/* tags with global attributes */
	       area, arrow, curve, eps, line, marker, text, xaxis, yaxis;
FILE *xml;


int main(int argc, char *argv[]) {
  int i, k, err;
  char xmlfile[GRX_PATHLENGTH], stylefile[GRX_PATHLENGTH], word[20], c;
  const struct tag_handle tagset[] = {
  /* tagname		function	adr. of global	attribute list */
    {" ",		NULL,		NULL,		NULL},
    {"?xml",		x_dummy,	NULL,		NULL},
    {"!DOCTYPE",	x_dummy,	NULL,		NULL},
    {"diagram",		x_diagram,	&diagram,	diagram_names},
    {"cartesian",	x_diagram,	&diagram,	diagram_names},
    {"area",		x_area,		&area,		area_names},
    {"arrow",		x_arrow,	&arrow,		arrow_names},
    {"bold",		x_bold,		NULL,		NULL},
    {"bolditalic",	x_bolditalic,	NULL,		NULL},
    {"comment",		x_comment,	NULL,		NULL},
    {"curve",		x_curve,	&curve,		curve_names},
    {"eps",		x_eps,		&eps,		eps_names},
    {"italic",		x_italic,	NULL,		NULL},
    {"line",		x_line,		&line,		line_names},
    {"marker",		x_marker,	&marker,	marker_names},
    {"parameters",	x_parameters,	NULL,		NULL},
    {"roman",		x_roman,	NULL,		NULL},
    {"sub",		x_sub,		NULL,		NULL},
    {"super",		x_super,	NULL,		NULL},
    {"symbol",		x_symbol,	NULL,		NULL},
    {"text",		x_text,		&text,		text_names},
    {"xaxis",		x_xaxis,	&xaxis,		xaxis_names},
    {"yaxis",		x_yaxis,	&yaxis,		yaxis_names},
    {"",		NULL, 		NULL,		NULL}
  };
  FILE *style;

  if (argc != 2) {
    fprintf(stderr, "usage: xml2eps DOC\n");
    fprintf(stderr, "  DOC is a directory with write permission\n");
    fprintf(stderr, "  Files DOC/diagram.xml and DOC/global.xml "
    "must exist.\n");
    return (GRX_FAIL);
  }

  for (i = 0; i < GRX_NPAR; i++) par[i] = 0.0;
  sprintf(xmlfile, "%s/diagram.xml", argv[1]);
  if ((xml = fopen(xmlfile, "r")) == NULL) {
    fprintf(stderr, "could not open diagram file %s\n", xmlfile);
    return (GRX_FAIL);
  }

  /* Global attributes are loaded in two stages: First we look up
     GRX_HOME/xml/cartesian.g.xml, where we keep defaults for all
     global attributes of all tags. Then we look up the diagram
     style file. If we add new tag definitions to Grafix at a
     later time, we do not have to update the style files of all
     old diagrams, but merely cartesian.g.xml. */
  sprintf(stylefile, "%s/xml/cartesian.g.xml", GRX_HOME);
  if ((style = fopen(stylefile, "r")) == NULL) {
    fprintf(stderr, "could not open style file %s\n", stylefile);
    return (GRX_FAIL);
  }
  if ((err = parse_global(style, tagset)) != 0) {
    fprintf(stderr, "loading of default global attributes failed\n");
    return (GRX_FAIL);
  }
  fclose(style);

  sprintf(stylefile, "%s/global.xml", argv[1]);
  if ((style = fopen(stylefile, "r")) == NULL) {
    fprintf(stderr, "could not open style file %s\n", stylefile);
    return (GRX_FAIL);
  }
  if ((err = parse_global(style, tagset)) != 0) {
    fprintf(stderr, "loading of global attributes failed\n");
    return (GRX_FAIL);
  }
  fclose(style);

  /* XML file processing: skip over XML preamble, then interpret
     the diagram tag pair (which contains the whole diagram) */
  while (1) {
    if ((err = fscanf(xml, "<%s", word)) != 1) {
      fprintf(stderr, "error %d: diagram.xml does not start with a tag\n",
      GRX_NOT_XML);
      return (GRX_NOT_XML);
    }
    if (!strcmp(word, "!DOCTYPE")) break;
    while ((c = getc(xml)) != '\n');	/* discard end-of-line */
  }
  if ((err = fscanf(xml, "%s", word)) != 1) {
    fprintf(stderr, "error %d: no DOCTYPE declaration\n", GRX_NOT_XML);
    return (GRX_NOT_XML);
  }
  while ((c = getc(xml)) != '\n');	/* discard end-of-line */
  if ((err = parse_xml(xml, 0, tagset)) != 0) fprintf(stderr,
  "finished with error %d\n", err);
  return (err);
}

/*************************************************************24.03.2007*/
/* parse_xml: identify XML tags					*/
/************************************************************************/
/* xml		(i)  XML input file
   parent_id	(i)  number of the tag which invoked this function
   tagset	(i)  array of defined tags
   RESULT	(o)  error code: 0 = no error				*/
/*======================================================================*/
int parse_xml(FILE *xml, int parent_id, const struct tag_handle *tagset) {
  int c, i, tag_id, err;
  char buffer[GRX_MAXBUF], *attr;

  while ((c = getc(xml)) != EOF) {
    if (c >= 0xe2) {			/* UTF-8 3-byte character */
      buffer[0] = c;
      buffer[1] = getc(xml);
      buffer[2] = getc(xml);
      buffer[3] = '\0';
      printf("%s", utf2ps3(buffer));
      continue;
    }
    if (c >= 0xc2) {			/* UTF-8 2-byte character */
      buffer[0] = c;
      buffer[1] = getc(xml);
      buffer[2] = '\0';
      printf("%s", utf2ps2(buffer));
      continue;
    }
    if (c > 0x7f) continue;		/* skip over non-ASCII chars */
    switch (c) {
    case '&': 				/* special character */
      switch (interpretation_mode) {
      case XML:
	fscanf(xml, "%[^;]%*c", buffer);
	print_special(buffer);
	break;
      case PS:
	putchar(c);
	break;
      default:;
      }
      break;
    case '<':				/* tag */
      fscanf(xml, "%[^>]%*c", buffer);
      if (buffer[0] == '?' | buffer[0] == '!') break;
      i = strlen(buffer) - 1;
      if (buffer[i] == '/') 		/* empty tag? */
        buffer[i] = '\0';
      strcat(buffer, " ");		/* we do not expect <xyz/> tags */
      if ((tag_id = recognize(buffer, tagset, &attr)) == 0)
      return (GRX_UNKNOWN_TAG);
      if (tag_id == -parent_id) return (GRX_GOOD);
      if (tag_id < 0) return (GRX_ENDTAG);
      if ((err = tagset[tag_id].process(xml, tag_id, tagset, attr,
      tagset[tag_id].global)) != 0) return (err);
      break;
    case '(':
      if (interpretation_mode == XML)
	printf("\\050");
      else if (interpretation_mode == PS)
	putchar(c);
      break;
    case ')':
      if (interpretation_mode == XML)
	printf("\\051");
      else if (interpretation_mode == PS)
	putchar(c);
      break;
    case '\n':				/* line feed */
      if (parent_id == 0) return (GRX_GOOD);
      if (interpretation_mode != COMMENT) putchar(c);
      break;
    default: 				/* plain text */
      if (c && interpretation_mode != COMMENT) putchar(c);
    }
  }

  return (GRX_EOF);			/* we should never come here */
}

/*************************************************************04.04.2002*/
/* recognize: identify a tag						*/
/************************************************************************/
/* tag		(i)  name of tag to be identified
   tagset	(i)  set of tag identifiers (must end with empty one)
   attr		(o)  pointer to attribute part of tag, NULL if tag
		     has no local modifications
   RESULT	(o)  number of tag in set; 0 denotes failure,
		     a negative number denotes a closing tag.		*/
/*======================================================================*/
int recognize(const char *tag, const struct tag_handle *tagset,
char **attr) {
  int i, closeflag, b;
  char *p, aux[GRX_MAXBUF];

  sscanf(tag, "%s%n", aux, &b);		/* extract tag name */
  i = 1;
  p = aux;
  if (*p == '/') {			/* is it a closing tag? */
    p++;
    closeflag = -1;
  }
  else
    closeflag = 1;

  *attr = (b == strlen(tag)) ? NULL : (char *) tag + b + 1;
     
  while (strcmp(tagset[i].name, "") != 0) {
    if (!strcmp(tagset[i].name, p)) return (i * closeflag);
    i++;
  }
  return (0);
}

/************************************************************************/
/* identify: locate a string in a list (ignore cases)			*/
/************************************************************************/
/* s		(i)  string to be identified
   slist	(i)  list of strings (terminated by an empty string)
   RESULT	(o)  index of string in list; -1 indicates failure	*/
/*======================================================================*/
int identify(const char *s, const char *slist[]) {
  int i;

  i = 0;
  while (*(slist[i]) != '\0') {
    if (strcasecmp(s, slist[i]) == 0) return (i);
    i++;
  }
  return (-1);
}
  
/*************************************************************11.01.2004*/
/* psunits: convert arbitrary units to PostScript points		*/
/************************************************************************/
/* s		(i)  some length, e.g., "5cm"
   RESULT	(i)  PostScript points					*/
/*======================================================================*/
double psunits(const char *s) {
  double x, c0, c1, g0, g1, pt;
  char unit[GRX_RLEN], aux[GRX_RLEN];

  *unit = '\0';
  sscanf(s, "%lf%s", &x, unit);
  if (!strcmp(unit, "cm"))
    pt = x * 72 / 2.54;
  else if (!strcmp(unit, "mm"))
    pt = x * 72 / 25.4;
  else if (!strcmp(unit, "in"))
    pt = x * 72 + 0.5;
  else if (!strcmp(unit, "pc"))
    pt = x * 12;
  else if (!strcmp(unit, "pt"))
    pt = x;
  else if (!strcmp(unit, "ru") || *unit == '\0')
    pt =  x * (psunits(diagram.att[3]) - psunits(diagram.att[1]));
  else if (!strcmp(unit, "xu")) {
    sscanf(diagram.att[9], "%lf", &c0);
    sscanf(diagram.att[11], "%lf", &c1);
    sscanf(diagram.att[5], "%lf", &g0);
    sscanf(diagram.att[7], "%lf", &g1);
    sprintf(aux, "%fru", x * (g1 - g0) / (c1 -c0));
    pt = psunits(aux);
  }
  else if (!strcmp(unit, "yu")) {
    sscanf(diagram.att[10], "%lf", &c0);
    sscanf(diagram.att[12], "%lf", &c1);
    sscanf(diagram.att[5], "%lf", &g0);
    sscanf(diagram.att[7], "%lf", &g1);
    sprintf(aux, "%fru", x * (g1 - g0) / (c1 -c0));
    pt = psunits(aux);
  }
  else
    pt = 0.0;
  return (pt);
}

/*************************************************************27.04.2002*/
/* parse_attlist: parse an attribute list, assign values to tag		*/
/*		  attributes						*/
/************************************************************************/
/* attstring	(i)  string containing attributes
		     format: name1="val1" name2="val2" ...
		     (last character must be a space)
   attnames	(i)  array of attribute names
   a		(io) attribute list
   RESULT	(o)  number of attributes recognized in attstring	*/
/*======================================================================*/
int parse_attlist(const char *attstring, const char *attnames[],
struct attlist *a) {
  int i, k, n, err;
  char *p, keyword[GRX_ALEN], keyval[GRX_RLEN], aux[GRX_ALEN + GRX_RLEN + 4];

  /* Attribute strings look like: keyword1="value1" keyword2=value2" . */
  n = 0;
  p = (char *) attstring;
  while (strlen(p) > 0) {
  while (*p <= 0x20) p++;		/* skip over CR, LF, TAB, etc. */
    sscanf(p, "%[^=]%*c%*c%n", keyword, &i);
    p += i;
    if (sscanf(p, "%[^\"]%*c%*c%n", keyval, &i) == 1)
      p += i;				/* value recognized */
    else {
      p += 2;				/* empty value */
      strcpy(keyval, "");
    }
    if ((k = identify(keyword, attnames)) >= 0) {
      strcpy(a->att[k], keyval);
      n++;
    }
  }
  return (n);
}

/*************************************************************08.05.2002*/
/* parse_global: load global attributes from style file			*/
/************************************************************************/
/* style	(i)  global style file
   tagset	(i)  array of available tags
   RESULT	(o)  error code: 0 = no error				*/
/*======================================================================*/
int parse_global(FILE *style, const struct tag_handle *tagset) {
  int i, tag_id;
  char c, buffer[GRX_ATTLEN], *attr;

  while ((c = getc(style)) != EOF) {
    if (c == '<') {
      fscanf(style, "%[^>]%*c", buffer);
      i = strlen(buffer) - 1;
      if (buffer[i] == '/') buffer[i] = '\0';
      strcat(buffer, " ");
      tag_id = recognize(buffer, tagset, &attr);
      if (tag_id <= 0 && tag_id != -4) {/* tolerate a </CLASS> tag */
      	fprintf(stderr, "xml2eps: unknown tag <%s>\n", buffer);
	return (GRX_UNKNOWN_TAG);
      }
      if (tagset[tag_id].global != NULL) {
        parse_attlist(attr, tagset[tag_id].attnames, tagset[tag_id].global);
      }
    }
  }
  return (GRX_GOOD);
}
